;
; File......: INT86.ASM
; Author....: Ted Means
; Date......: $Date:   02 Apr 1991 18:24:54  $
; Revision..: $Revision:   1.0  $
; Log file..: $Logfile:   E:/nanfor/src/int86.asv  $
; 
; This function is an original work by Ted Means and is placed in the
; public domain.
;
; Modification history:
; ---------------------
;
; $Log:   E:/nanfor/src/int86.asv  $
;  
;     Rev 1.0   02 Apr 1991 18:24:54   GLENN
;  Nanforum Toolkit
;
;


;  $DOC$
;  $FUNCNAME$
;      FT_INT86()
;  $CATEGORY$
;      DOS/BIOS
;  $ONELINER$
;      Execute a software interrupt
;  $SYNTAX$
;      FT_INT86( nInterruptNumber, aRegisterValues ) -> nResult
;  $ARGUMENTS$
;      <nInterruptNumber> is the interrupt to execute.
;
;      <aRegisterValues> is an array that contains values to be loaded
;      into the various CPU registers.  The correspondence between
;      registers and array elements is as follows:
;
;               aElement[1]  ==  AX register
;               aElement[2]  ==  BX register
;               aElement[3]  ==  CX register
;               aElement[4]  ==  DX register
;               aElement[5]  ==  SI register
;               aElement[6]  ==  DI register
;               aElement[7]  ==  BP register
;               aElement[8]  ==  DS register
;               aElement[9]  ==  ES register
;               aElement[10] ==  Flags register
;               aElement[11] ==  String parameter for DS register
;               aElement[12] ==  String parameter for ES register
;  $RETURNS$
;      0 if all parameters valid and the function was able
;        to execute the desired interrupt.
;     -1 if invalid parameters passed.
;     -2 if the parameters were okay, but the interrupt was not
;        called because the memory required for a string parameter
;        could not be allocated.
;
;     In addition, the array elements will contain whatever values were in
;     the CPU registers immediately after the interrupt was executed.  If
;     either of the string parameters were altered by the interrupt, these
;     changes will be reflected as well.
;
;  $DESCRIPTION$
;     It is occassionally useful to be able to call interrupts directly from
;     Clipper, without having to write a separate routine in C or ASM.  This
;     function allows you that capability.
;
;     Given Clipper's high-level orientation, this function is necessarily
;     somewhat messy to use.  First, declare an array of twelve elements to
;     hold all the values for the CPU registers and two string parameters.
;     Then initialize the array elements with the values that you want the
;     CPU registers to contain when the interrupt is executed.  You need not
;     initialize all the elements; for example, if the interrupt requires
;     you to specify values for AX, DX, and DS, you would only need to
;     initialize elements 1, 4, and 8.
;
;     Once you have done the required register setup, call FT_INT86(),
;     passing the interrupt number and the register array as parameters.
;     The function will load the CPU with your specified values, execute the
;     interrupt, and then store the contents of the CPU registers back into
;     your array.  This will allow you to evaluate the results of the
;     interrupt.
;
;     Some interrupt services require you to pass the address of a string in
;     a pair of registers.  This function is capable of handling these sorts
;     of situations, but it will take a little work on your part.
;
;     As you may have noticed, in addition to the 10 array elements
;     corresponding to the CPU registers, there are a couple of slots for
;     string parameters.  If you need to pass a string that uses the DS
;     register, then use element 11; if you need to pass a string that uses
;     the ES register, use element 12.  FT_INT86() will detect that you've
;     supplied a string and will use the segment address of the string
;     instead of the values in the array.
;
;     That takes care of obtaining the segment portion of the pointer.  To
;     specify which register is to contain the offset, use the
;     pseudo-function Register() which is contained in the FTINT86.CH file.
;     Assign this function to array element 11 or 12, then concatenate your
;     string. FT_INT86() will examine the string and use the first byte to
;     determine which register will contain the offset.  If this sounds
;     confusing, look at the examples below; they make it much clearer than
;     I am able to explain it.
;
;     Note that some interrupts will change the value of the string.  This
;     function will take the appropriate steps to make sure the modified
;     string is stored back into element 11 or 12 of the array.  Before
;     doing so, it drops the first byte of the string, which was used to
;     specify the offset register.  Thus, upon return, only the actual
;     string will be in stored in element 11 or 12.
;
;     All the CPU registers are sixteen bits in size.  Some, however, are
;     also split into two 8-bit registers.  This function is only capable of
;     receiving and returning registers that are 16 bits in size.  To split
;     a 16-bit register into two 8-bit values, you can use the following use
;     the pseudo-functions contained in the .CH file.
;
;     To alter an 8-bit number so it will appear in the high-order byte of a
;     register when passed to the FT_INT86() function, use the MakeHI()
;     pseudo-function contained in the .CH file.
;
;     The source is written for Turbo Assembler's IDEAL mode.  To use
;     another assembler will require very minor changes to the syntax.
;
;  $EXAMPLES$
;
;     * This example shows how to call the DOS "create file" service.  Take
;     * special note of how to set up string parameters.
;
;     #include "FTINT86.CH"
;
;     local aRegs[12]              && Declare the register array
;     aRegs[ AX ] := makehi(60)    && DOS service, create file
;     aRegs[ CX ] := 0             && Specify file attribute
;
;     * Pay attention here, this is crucial.  Note how to set up the string
;     * so it appears in DS:DX.
;
;     aRegs[ DS_STRING ] := Register( DX ) + "C:\MISC\MYFILE.XXX"
;
;     FT_INT86( 33, aRegs)         && Make the call to the DOS interrupt
;
;
;     * This example shows how to call the DOS "get current directory"
;     * service.  This one also uses a string parameter, but note that it
;     * uses a different offset register.
;
;     #include "FTINT86.CH"
;
;     local aRegs[12]
;     aRegs[ AX ] := makehi(71)
;     aRegs[ DX ] := 0           && Choose default drive
;
;     * This service requires a 64-byte buffer whose address is in DS:SI.  DOS
;     * will fill the buffer with the current directory.
;
;     aRegs[ DS_STRING] := Register( SI ) + space(64)
;
;     FT_INT86( 33, aRegs)
;
;     ? aRegs[ DS_STRING ]       && Display the directory name
;
;
;
;     * For the sake of completeness, here's an example that doesn't use a
;     * string.  This one changes the video mode.
;
;     #include "FTINT86.CH"
;
;     local aRegs[12]
;
;     aRegs[ AX ] := 16          && Choose hi-res graphics
;     FT_INT86( 16, aRegs)
;  $INCLUDE$
;     FTINT86.CH
;  $END$

         IDEAL

Public   FT_Int86

Extrn    __ParInfo:Far
Extrn    __ParInfa:Far
Extrn    __ParNI:Far
Extrn    __ParC:Far
Extrn    __ParCLen:Far
Extrn    __StorNI:Far
Extrn    __StorCLen:Far
Extrn    __RetNI:Far
Extrn    __XAlloc:Far
Extrn    __XFree:Far
Extrn    __XUnlock:Far

Segment  _NanFor   Word      "CODE"
         Assume    CS:_NanFor

; The following data is stored in the code segment, but don't panic.  It
; won't render this function unoverlayable.  This function doesn't expect
; any of this data to be maintained in between calls, so overlaying it is
; okay.
;
; Storing it in the code segment is necessary because ES and DS will both be
; loaded with values passed from the caller, so this local data would be
; unaccessible unless the code segment is used for storage.

Label    IntAddr   DWord                     ; This will contain the address
IntOfs   DW        ?                         ; of the desired interrupt
IntSeg   DW        ?                         ; handler

RegAX    DW        ?                         ; Supplied register values
RegBX    DW        ?
RegCX    DW        ?
RegDX    DW        ?
RegSI    DW        ?
RegDI    DW        ?
RegBP    DW        ?
RegDS    DW        ?
RegES    DW        ?
RegFlags DW        ?

IntNo    DB        ?                         ; The interrupt that was requested

Label    DSPtr     DWord                     ; Address of string using DS
DSOfs    DW        ?
DSSeg    DW        ?

Label    ESPtr     DWord                     ; Address of string using ES
ESOfs    DW        ?
ESSeg    DW        ?

DSLen    DW        ?                         ; Length of string using DS
ESLen    DW        ?                         ; Length of string using ES

RetVal   DW        ?                         ; Function's return value

Proc     FT_Int86  Far

         Mov       [CS:DSSeg],0              ; Nullify pointers
         Mov       [CS:DSOfs],0
         Mov       [CS:ESSeg],0
         Mov       [CS:ESOfs],0
         Mov       [CS:RetVal],0             ; Intialize return value

         Xor       AX,AX                     ; Request param count
         Push      AX                        ; Put request on stack
         Call      __ParInfo                 ; Get param count
         Add       SP,2                      ; Realign stack
         Cmp       AX,2                      ; At least 2 params?
         JNB       GetType1                  ; If so, continue

BadParam:Mov       [CS:RetVal],-1            ; Set return value to -1
         Jmp       Exit                      ; No, so quit

GetType1:Mov       AX,1                      ; Specify first param
         Push      AX                        ; Put param # on stack
         Call      __ParInfo                 ; Get param type
         Add       SP,2                      ; Realign stack
         Test      AX,2                      ; Numeric?
         JZ        BadParam                  ; No, so quit

GetType2:Mov       AX,2                      ; Specify second param
         Push      AX                        ; Put param # on stack
         Call      __ParInfo                 ; Get param type
         Add       SP,2                      ; Realign stack
         Test      AX,512                    ; Array?
         JZ        BadParam                  ; No, so quit

         Xor       AX,AX                     ; Request element count
         Push      AX                        ; Put request on stack
         Mov       AX,2                      ; Specify second param
         Push      AX                        ; Put param # on stack
         Call      __ParInfa                 ; Get element count
         Add       SP,4                      ; Realign stack
         Cmp       AX,12                     ; 12 elements?
         JNE       BadParam                  ; If not, exit

         Mov       BX,Offset _NanFor:RegES   ; Offset of ES register value
         Mov       CX,9                      ; Loop till ninth element

GetRegs: Push      BX                        ; Save offset
         Push      CX                        ; Save element # on stack
         Mov       AX,2                      ; Specify second param
         Push      AX                        ; Put param # on stack
         Call      __ParNI                   ; Get value
         Add       SP,2                      ; Realign stack
         Pop       CX                        ; Restore element #
         Pop       BX                        ; Restore BX
         Mov       [Word Ptr CS:BX],AX       ; Save supplied register value
         Sub       BX,2                      ; Point BX to next value
         Loop      GetRegs                   ; Get next element value

         Mov       AX,Offset ESString        ; Alternate return address
         Push      AX                        ; Put on stack
         Mov       AX,11                     ; Specify element #
         Call      Near _FTCopy              ; Call copy routine
         JC        NoAlloc                   ; If error, report it
         Add       SP,2                      ; Realign stack
         Dec       AX                        ; Adjust for zero offset
         And       AX,111b                   ; Mask unwanted bits
         SHL       AX,1                      ; Multiply by 2
         Mov       BX,Offset _NanFor:RegAX   ; Get pointer to first register
         Add       BX,AX                     ; Point to requested register
         Mov       AX,[CS:DSOfs]             ; Get offset value
         Mov       [Word Ptr CS:BX],AX       ; Put in register container
         Mov       AX,[CS:DSSeg]             ; Get segment value
         Mov       [CS:RegDS],AX             ; Put in register container
         Jmp       Short ESString            ; Do ES now

NoAlloc: Add       SP,2                      ; Realign stack
         Mov       [CS:RetVal],-2            ; Indicate allocation failure
         Jmp       CheckDS                   ; Quit

ESString:Mov       AX,Offset GetInt          ; Alternate return address
         Push      AX                        ; Put on stack
         Mov       AX,12                     ; Specify element #
         Call      Near _FTCopy              ; Call copy routine
         JC        NoAlloc                   ; If error, report it
         Add       SP,2                      ; Realign stack
         Dec       AX                        ; Adjust for zero offset
         And       AX,111b                   ; Mask unwanted bits
         SHL       AX,1                      ; Multiply by 2
         Mov       BX,Offset _NanFor:RegAX   ; Get pointer to first register
         Add       BX,AX                     ; Point to requested register
         Mov       AX,[CS:ESOfs]             ; Get offset value
         Mov       [Word Ptr CS:BX],AX       ; Put in register container
         Mov       AX,[CS:ESSeg]             ; Get segment value
         Mov       [CS:RegES],AX             ; Put in register container

GetInt:  Mov       AX,1                      ; Specify first param
         Push      AX                        ; Put param # on stack
         Call      __ParNI                   ; Get interrupt #
         Add       SP,2                      ; Realign stack
         Mov       [CS:IntNo],AL             ; Store interupt # for later
         Mov       AH,35h                    ; DOS service--get int vector
         Int       21h                       ; Call DOS
         Mov       [CS:IntSeg],ES            ; Get segment interrupt handler
         Mov       [CS:IntOfs],BX            ; Get offset of interrupt handler

         Push      DS                        ; Save critical registers
         Push      SI
         Push      DI
         Push      BP

         Mov       CX,9                      ; Specify loop value
         Mov       BX,Offset _NanFor:RegAX   ; Offset of register values

ToStack: Push      [Word Ptr CS:BX]          ; Put register values on stack
         Add       BX,2                      ; Point BX to next value
         Loop      ToStack                   ; Do the next one
         Pop       ES                        ; Now load registers
         Pop       DS
         Pop       BP
         Pop       DI
         Pop       SI
         Pop       DX
         Pop       CX
         Pop       BX
         Pop       AX
         PushF                               ; Simulate interrupt
         CLI
         Call      [CS:IntAddr]              ; Call interrupt handler

         Mov       [CS:RegAX],AX             ; Store values
         Mov       [CS:RegBX],BX
         Mov       [CS:RegCX],CX
         Mov       [CS:RegDX],DX
         Mov       [CS:RegSI],SI
         Mov       [CS:RegDI],DI
         Mov       [CS:RegBP],BP
         Mov       [CS:RegDS],DS
         Mov       [CS:RegES],ES
         Pushf
         Pop       [CS:RegFlags]

         Cmp       [CS:IntNo],25h            ; Was it INT 25h?
         JB        PopRegs                   ; No, so don't worry
         Cmp       [CS:IntNo],26h            ; Was it INT 26h?
         JA        PopRegs                   ; No, so don't worry
         Add       SP,2                      ; 25h and 26h don't clean up stack

PopRegs: Pop       BP                        ; Restore critical regs
         Pop       DI
         Pop       SI
         Pop       DS

         Mov       CX,10                     ; Total # of elements
         Mov       BX,Offset _NanFor:RegFlags; Get address of last element
PutRegs: Push      BX                        ; Save BX
         Push      CX                        ; Put element # on stack
         Mov       AX,2                      ; Specify param # 2
         Push      AX                        ; Put param # on stack
         Push      [Word Ptr CS:BX]          ; Put register value on stack
         Call      __StorNI                  ; Put value in array element
         Add       SP,4                      ; Realign stack
         Pop       CX                        ; Restore CX
         Pop       BX                        ; Restore BX
         Sub       BX,2                      ; Point BX to next element
         Loop      PutRegs                   ; Do the next value
         Xor       AX,AX                     ; Set return value to zero

CheckDS: Mov       DX,[CS:DSSeg]             ; Load segment portion
         Mov       AX,[CS:DSOfs]             ; Load offset portion
         Test      DX,0FFFFh                 ; Test for null pointer
         JNZ       StoreDS                   ; If not, continue
         Test      AX,0FFFFh                 ; Test for null pointer
         JNZ       StoreDS                   ; If not, continue
         Jmp       Short CheckES             ; Null pointer, so don't dealloc

StoreDS: Mov       CX,[CS:DSLen]             ; Get length
         Mov       BX,11                     ; Element #
         Push      BX                        ; Put element # on stack
         Mov       BX,2                      ; Param #
         Push      BX                        ; Put param # on stack
         Push      CX                        ; Put length on stack
         Push      DX                        ; Put segment on stack
         Push      AX                        ; Put offset on stack
         Call      __StorCLen                ; Assign string
         Call      __XFree                   ; Free allocated RAM
         Add       SP,10                     ; Realign stack

CheckES: Mov       DX,[CS:ESSeg]             ; Load segment portion
         Mov       AX,[CS:ESOfs]             ; Load offset portion
         Test      DX,0FFFFh                 ; Null pointer?
         JNZ       StoreES                   ; No, continue
         Test      AX,0FFFFh                 ; Null pointer?
         JNZ       StoreES                   ; No, continue
         Jmp       Short Exit                ; Null pointer, so don't dealloc

StoreES: Mov       CX,[CS:ESLen]             ; Get length
         Mov       BX,12                     ; Element #
         Push      BX                        ; Put element # on stack
         Mov       BX,2                      ; Param #
         Push      BX                        ; Put param # on stack
         Push      CX                        ; Put length on stack
         Push      DX                        ; Put segment on stack
         Push      AX                        ; Put offset on stack
         Call      __StorCLen                ; Assign string
         Call      __XFree                   ; Free allocated RAM
         Add       SP,10                     ; Realign stack

Exit:    Push      [CS:RetVal]               ; Put return value on stack
         Call      __RetNI                   ; Return it to Clipper app
         Add       SP,2                      ; Realign stack

         Ret
Endp     FT_Int86



Proc     _FTCopy   Near

         Push      AX                        ; Put element # on stack
         Mov       AX,2                      ; Specify param #2
         Push      AX                        ; Put param # on stack
         Call      __ParInfa                 ; Get data type
         Test      AX,1                      ; Character?
         JNZ       GetLen                    ; Yes, so continue
         Add       SP,6                      ; Clear stack
         Ret                                 ; Using alternate return address

GetLen:  Call      __ParCLen                 ; Get length
         Dec       AX                        ; Adjust for offset marker
         Add       SP,2                      ; Realign stack
         Pop       DX                        ; Get element # back
         Push      DX                        ; Put back for later
         Cmp       DX,11                     ; DS reg?
         JNE       LenES                     ; Nope, so it's ES
         Mov       [CS:DSLen],AX             ; Store length
         Jmp       Short GetRAM              ; Continue

LenES:   Mov       [CS:ESLen],AX             ; Store length

GetRAM:  Push      AX                        ; Put length on stack
         Call      __XAlloc                  ; Allocate some RAM
         Add       SP,2                      ; Realign stack
         Test      DX,0FFFFh                 ; Check for null pointer
         JNZ       PtrDS                     ; If not, continue
         Test      AX,0FFFFh                 ; Check for null pointer
         JNZ       PtrDS                     ; If not, so continue
         Add       SP,2                      ; Clear stack
         STC                                 ; Indicate allocation failed
         Ret

PtrDS:   Pop       CX                        ; Get element back
         Cmp       CX,11                     ; DS Reg?
         JNE       PtrES                     ; Nope, so use ES
         Mov       [CS:DSSeg],DX             ; Store segment of allocated RAM
         Mov       [CS:DSOfs],AX             ; Store offset of allocated RAM
         Jmp       Short GetPtr              ; Get passed string

PtrES:   Mov       [CS:ESSeg],DX             ; Store segment of allocated RAM
         Mov       [CS:ESOfs],AX             ; Store offset of allocated RAM

GetPtr:  Push      DS                        ; Save registers
         Push      SI
         Push      DI

         Push      CX                        ; Put element # on stack
         Mov       AX,2                      ; Specify param #
         Push      AX                        ; Put on stack
         Call      __ParC                    ; Get pointer
         Add       SP,2                      ; Realign stack
         Pop       CX                        ; Get element # back
         Cmp       CX,11                     ; DS reg?
         JNE       LoadES                    ; Nope, so use ES
         LES       DI,[DWord Ptr CS:DSPtr]   ; Get pointer to local buffer
         Mov       CX,[CS:DSLen]             ; Get length
         Jmp       Short MakeCopy            ; Copy the string

LoadES:  LES       DI,[DWord Ptr CS:ESPtr]   ; Get pointer to local buffer
         Mov       CX,[CS:ESLen]             ; Get length

MakeCopy:Mov       DS,DX                     ; Load segment portion
         Mov       SI,AX                     ; Load offset portion
         CLD                                 ; Make sure we go forward
         Lodsb                               ; Get register indicator
         Rep       Movsb                     ; Copy the string
         Pop       DI                        ; Restore registers
         Pop       SI
         Pop       DS
         Push      AX                        ; Save offset register value
         Call      __XUnlock                 ; Tell VMM to let go
         Pop       AX                        ; Restore AX
         Ret
Endp     _FTCopy

Ends     _NanFor
End

