;
; File......: SHADOW.ASM
; Author....: Reginald B. Walton
; Date......: $Date:   02 Apr 1991 18:25:40  $
; Revision..: $Revision:   1.0  $
; Log file..: $Logfile:   E:/nanfor/src/shadow.asv  $
; 
; This is an original work by Reginald B. Walton and is placed in the
; public domain.
;
; Modification history:
; ---------------------
;
; $Log:   E:/nanfor/src/shadow.asv  $
;  
;     Rev 1.0   02 Apr 1991 18:25:40   GLENN
;  Nanforum Toolkit
;
;


;  $DOC$
;  $FUNCNAME$
;      FT_SHADOW()
;  $CATEGORY$
;      Video
;  $ONELINER$
;      Draw a non-destructive shadow on the screen
;  $SYNTAX$
;     FT_SHADOW( <nTop>, <nLeft>, <nBottom>, <nRight> ) -> lResult
;  $ARGUMENTS$
;     <nTop>    is the top row of the shadow area.
;
;     <nLeft>   is the upper left column of the shadow area.
;
;     <nBottom> is the bottom row of the shadow area.
;
;     <nRight>  is the lower right column of the shadow area.
;  $RETURNS$
;     .T. if the shadow was drawn.
;     .F. if the shadow was not drawn.
;  $DESCRIPTION$
;     This function was designed to have my application emulate the 
;     non-destructive shadowing of the PC-TOOLS application. This function
;     was written in Assembly language for speed since, the same function
;     written in Clipper was too slow. The size of this function could be
;     reduced by re-writing the SET_CUR_POSITION, READ_CHARACTER, and 
;     WRITE_CHARACTER macros as procedures. Note: this will slow it down
;     a tad, since stack manipulation will be involved.
;
;     1.  If you are using the SAVESCREEN() function to save your screen,
;         remember to save enough screen for the shadow. The shadow will
;         require 1 row and 2 columns.
;
;             save_win = SAVESCREEN(Trow,Tcol,Brow+1,Bcol+2)
;
;     2.  If you specify screen cordinated outside the 24 row or 80 column
;         boundry, no shadow will be drawn.  Thus, this function will only
;         work in standard text mode.
;          
;     3.  This function utilizes the standard IBM BIOS services to display
;         the shadow and could possibly create some problems in less-known
;         non-IBM compatibles.
;
;     The source code is written to MASM specifications.  To use another
;     assembler, either invoke that assembler's MASM compatibility switch,
;     or make the required modifications to the source code.
;  $EXAMPLES$
;     LOCAL Save_Win := SaveScreen(10,10,21,52)  // save enough for the shadow
;     FT_SHADOW(10,10,20,50)                     // draw shadow
;
;        ...
;          do what ya' like
;        ...
;
;     RestScreen(10,10,21,52,Save_Win)        // restore screen
;     RETURN
;  $END$


public   FT_SHADOW

INCLUDE  EXTENDA.MAC

TOPROW   EQU  BYTE PTR BP - 1              ; TOP ROW
TOPCOL   EQU  BYTE PTR BP - 2              ; TOP COLUMN
BOTROW   EQU  BYTE PTR BP - 3              ; BOTTOM ROW
BOTCOL   EQU  BYTE PTR BP - 4              ; BOTTOM COLUMN
CPAGE    EQU  BYTE PTR BP - 6              ; CURRENT PAGE

;
; THESE MACROS CAN BE RE-WRITTEN INTO PROCEDURES FOR SIZE DECREASE.
;
SET_CUR_POS MACRO
   MOV   AH,02H      ; SET CURSOR POSITION
   MOV   BH,[CPAGE]  ; CURRENT DISPLAY PAGE
   INT   10H         ; CALL BIOS
ENDM

READ_CHARACTER MACRO
   LOCAL OUTTA
   MOV   AH,08H         ; READ CHARACTER & ATTRIBUTE AT CURSOR
   MOV   BH,[CPAGE]     ; CURRENT DISPLAY PAGE
   INT   10H            ; CALL BIOS
                        ; AH = ATTRIBUTE
                        ; AL = CHARACTER

        ; USE THIS PORTION IF YOU HAVE DRAWN A PC-TOOLS LIKE DESKTOP
        ; ON YOUR SCREEN USING THE ASCII CHARACTER 178.
   CMP   AL,178         ; IF THE CHARACTER IS THE ASCII 178
   JNE   OUTTA          ;  MAKE IT THE ASCII 176 TO EMULATE
   SUB   AL,2           ;  A CONTRAST CHANGE.
   OUTTA:
ENDM

WRITE_CHARACTER   MACRO
   MOV   AH,09H         ; WRITE CHARACTER & ATTRIBUTE AT CURSOR
   MOV   BH,[CPAGE]     ; CURRENT DISPLAY PAGE
   MOV   BL,00000111B   ; BL = ATTRIBUTE (white/black)
   mov   cx,1           ; Times to write character is 1
   int   10h
endm

_nanfor  segment   word      public    'CODE'      ; declare code segment
         assume    cs:_nanfor

Ft_shadow          proc      far         ; start of routine

         push      bp             ; save bp
         mov       bp,sp          ; set up stack reference
         sub       sp,6           ; allocate room for locals

         get_pcount               ; How many parameters?
         cmp       ax,4           ; are there 4 parameters?
         jnb       allok          ; yes there are 4
         xor       ax,ax          ; clear ax register for .F. return code
         jmp       exit           ; return to clipper
 
allok:   get_int 1                ; Get the first parameter
         mov       [toprow],al    ; store it in TROW

         get_int 2                ; Get the second parameter
         mov       [topcol],al    ; store it in TCOL

         get_int 3                ; Get the third parameter
         mov       [botrow],al    ; store it in BROW

         get_int 4                ; get the fourth parameter
         mov       [botcol],al    ; store it in BCOL

get_current_page:                 ; get the current display page
         mov       ah,0fh         ; and save it for future reference
         int       10h
         mov       [cpage],bh
 
verti_shadow:                     ; draw vertical shadow
         mov       dh,[botrow]    ; start at the bottom row
         inc       dh             ;  plus 1

         cmp       dh,24          ; will the shadow be past the last row
         ja        hori_shadow    ; yes, then don't draw the verti shadow

         mov       dl,[topcol]    ; start at the top column
                                  ; increment the column inside loop
         xor       cx,cx          ; clear CX
         mov       cl,[botcol]    ; length of vertical shadow is 
         inc       cx             ;  the bottom column plus 1
         xor       ax,ax          ;  minus the top column
         mov       al,[topcol]
         sub       cx,ax

@loop1:                           ; This the loop for the vertical line
         push      ax             ; save registers
         push      bx
         push      cx
         inc       dl             ; position on next column
         push      dx             ; save current row and column

         set_cur_pos              ; set cursor position
         read_character           ; read character at position
         write_character          ; write new attribute at position

         pop       dx             ; restore registers
         pop       cx
         pop       bx
         pop       ax       
         loop      @loop1

Hori_shadow:                      ; draw horzontal shadow
         mov       dh,[toprow]    ; start at top row, increment inside loop
         mov       dl,[botcol]    ; start at bottom column 
         inc       dl             ; plus 1

         cmp       dl,80          ; will the shadow be past the last column
         jae       exit           ; yes, then don't draw the verti shadow

         xor       cx,cx          ; clear CX
         mov       cl,[botrow]    ; length of hori shadow is 
         inc       cx             ; the bottom row plus 1
         xor       ax,ax          ;  minus the top row
         mov       al,[toprow]         
         sub       cx,ax
 
@loop2:                           ; This is the loop for the the hori line
         push      ax             ; save registers
         push      bx
         push      cx
         inc       dh             ; position on next row
         push      dx             ; save current row and column

         set_cur_pos              ; set cursor position
         read_character           ; read character at position
         write_character          ; write new attribute at position

         inc       dl                ; next column
         cmp       dl,80             ; will the shadow be past the last column
         jae       past_second_verti ; don't draw the second verti shadow

         set_cur_pos              ; set cursor position
         read_character           ; read character at position
         write_character          ; write new attribute at position

past_second_verti:
         pop       dx             ; restore registers
         pop       cx
         pop       bx
         pop       ax       

         loop      @loop2
         mov       ax,1                   ; set return code to .T.

exit:    mov       sp,bp
         pop       bp
         ret_logical    ax          ; Return the value in AX
         ret                        ; Go on back to Clipper

Ft_shadow          endp

_nanfor            ends
end

